import argparse
import logging
import os
import signal
import sys
from typing import Any, Callable, Dict, Generic, List, Literal, Optional, TextIO, Type, TypeVar, Sequence, Union, cast, overload

from . import config
from .formatter import ColorFormatter

try:
    from argcomplete import autocomplete  # type: ignore[import]
except ImportError:
    def autocomplete(parser: argparse.ArgumentParser) -> None:
        ...


# pre-exec section
DEFAULT_LOG_LEVEL = logging.root.level
if ("--debug" in sys.argv) or ("-d" in sys.argv):
    DEFAULT_LOG_LEVEL = logging.DEBUG
elif ("--verbose" in sys.argv) or ("-v" in sys.argv):
    DEFAULT_LOG_LEVEL = logging.INFO


class Namespace(argparse.Namespace):
    prog: str
    version: str
    debug: bool
    _verbose: bool
    output: Optional[TextIO]

    @property
    def verbose(self) -> bool:
        return self.debug or self._verbose


N = TypeVar("N", bound=Namespace)
_N = TypeVar("_N")


class ArgumentParser(argparse.ArgumentParser, Generic[N]):
    NAME: Optional[str] = None
    DESCRIPTION: Optional[str] = None
    VERSION: Optional[str] = None
    NAMESPACE_TYPE: Type[N]
    LOGGING_FORMATTER: logging.Formatter = ColorFormatter()

    def __init__(
            self,
            prog: Optional[str] = None,
            version: Optional[str] = None,
            description: Optional[str] = None,
            epilog: Optional[str] = None,
            usage: Optional[str]=None,
            parents: List[argparse.ArgumentParser] = [],
            prefix_chars: str = "-",
            fromfile_prefix_chars: Optional[str] = None,
            argument_default: Any = None,
            conflict_handler: Literal["error", "resolve"] = "error",
            add_help: bool = True,
            allow_abbrev: bool = True,
            logger: Optional[logging.Logger] = None,
            formatter: Optional[logging.Formatter] = None) -> None:
        assert hasattr(self, "NAMESPACE_TYPE"), f"namespace type NAMESPACE_TYPE of '{self.__class__.__name__}' must be set"
        self.version = version or self.VERSION
        assert self.version is not None
        self.logger = logger or logging.getLogger()
        self.logging_formatter = formatter or self.LOGGING_FORMATTER
        super().__init__(
            prog=self.NAME or prog,
            usage=usage,
            description=description or self.DESCRIPTION,
            epilog=epilog,
            parents=parents,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            prefix_chars=prefix_chars,
            fromfile_prefix_chars=fromfile_prefix_chars,
            argument_default=argument_default,
            conflict_handler=conflict_handler,
            add_help=add_help,
            allow_abbrev=allow_abbrev)

    def error(self, message: str) -> Any:
        self.print_usage(sys.stderr)
        raise argparse.ArgumentTypeError(message)

    def __init_args__(self) -> None:
        self.add_argument("--version", action="version", version=f"%(prog)s {self.version}")
        verbosity = self.add_mutually_exclusive_group()
        verbosity.add_argument(
                "-v",
                "--verbose",
                action="store_true",
                dest="_verbose",
                help="enable verbose output")
        verbosity.add_argument(
                "-d",
                "--debug",
                action="store_true",
                help="enable debug output")
        self.add_argument(
                "-o",
                "--log-output",
                type=argparse.FileType("w"),
                default=None,
                dest="output",
                help="logging destination, default is stderr")

    @overload
    def parse_args(self, args: Optional[Sequence[str]] = ...) -> N:
        ...
    @overload
    def parse_args(self, args: Optional[Sequence[str]], namespace: None) -> N:
        ...
    @overload
    def parse_args(self, args: Optional[Sequence[str]], namespace: _N) -> _N:
        ...
    @overload
    def parse_args(self, *, namespace: None) -> N:
        ...
    @overload
    def parse_args(self, *, namespace: _N) -> _N:
        ...

    def parse_args(self, args: Optional[Sequence[str]] = None, namespace: Optional[_N] = None) -> _N:
        # set guessed logging level and add preliminary logger
        logger = self.logger
        log_level = logger.level
        logger.setLevel(DEFAULT_LOG_LEVEL)
        sh = logging.StreamHandler()
        sh.setFormatter(self.logging_formatter)
        logger.addHandler(sh)

        # initialize additional arguments
        self.__init_args__()
        # execute autocomplete
        autocomplete(self)

        if namespace is None:
            namespace = cast(_N, self.NAMESPACE_TYPE())
        setattr(namespace, "prog", self.prog)
        setattr(namespace, "version", self.version)
        argns = super().parse_args(args, namespace)

        # update logging level
        if getattr(argns, "debug", False):
            log_level = logging.DEBUG
        elif getattr(argns, "_verbose", False):
            log_level = logging.INFO
        logger.setLevel(log_level)

        # update logging handler
        logfile: Optional[TextIO] = getattr(argns, "output", None)
        if logfile is not None:
            logger.removeHandler(sh)
            sh = logging.StreamHandler(logfile)
            sh.setFormatter(self.logging_formatter)
            logger.addHandler(sh)
        return argns

    def exec(self,
            main: Callable[[N], Optional[int]],
            args: Optional[Sequence[str]] = None,
            namespace: Optional[N] = None) -> None:
        logger = self.logger
        try:
            argns: N = self.parse_args(args, namespace)
            logger.debug("running %s %s", self.prog, self.version)
            exit(main(argns) or 0)
        except Exception as exc:
            if logger.level <= logging.DEBUG:
                logger.exception(exc)
            else:
                if len(exc.args) == 0:
                    fallback = True
                    for texc, msg in config.EXCEPTION_INFORMATION.items():
                        if isinstance(exc, texc):
                            logger.error("%s: %s", self.prog, msg)
                            fallback = False
                    if fallback:
                        name = type(exc).__name__
                        name = name[name.rfind(".")+1:]
                        logger.error("%s: %s", self.prog, name)
                else:
                    logger.error("%s: %s", self.prog, ", ".join(exc.args))
            if isinstance(exc, argparse.ArgumentTypeError):
                exit(2)
            exit(1)
        except KeyboardInterrupt as kint:
            print(flush=True)
            logger.debug("%s: keyboard interrupt", argns.prog)
            signal.signal(signal.SIGINT, signal.SIG_DFL)
            os.kill(os.getpid(), signal.SIGINT)




import sys

from typing import Union, Optional, Tuple

from ..fmt import Text
from . import config


def print_question(text: Union[str, Text], inline: Optional[bool] = None) -> None:
    if inline is None:
        inline = len(Text(text).unformatted) <= config.AUTO_INLINE_MAX_CHARACTER_COUNT
    txt: Tuple[Union[str, Text], ...]
    if inline:
        txt = (config.QUESTION_PREFIX, text, config.QUESTION_INLINE_SUFFIX,)
    else:
        txt = (text, "\r\n", config.QUESTION_PREFIX,)
    print(Text(*txt, config=config.QUESTION_FORMAT), end="", flush=True)




from typing import Callable, Optional, TypeVar

from argparse import ArgumentTypeError

from . import unix

T = TypeVar("T")

def select_char(chars: str, ignore_case: bool = True, allow_empty: bool = True, silent: bool = False) -> Optional[str]:
    return unix.select_char(chars, ignore_case, allow_empty, silent)

def select_bool(default: Optional[bool] = None, ignore_case: bool = True, allow_empty: bool = True, silent: bool = False) -> Optional[bool]:
    result = select_char("yn", ignore_case, allow_empty, silent)
    return (result.lower() == "y") if result else default

def select_str(verify: Callable[[str], bool] = lambda text: True, default: Optional[str] = None, silent: bool = False) -> str:
    result = unix.select_str(lambda text: verify(text) or (len(text) == 0 and default is not None), silent)
    return result if result or not default else default

def select(convert: Callable[[str], T], default: Optional[T] = None, allow_empty: bool = False, silent: bool = False) -> Optional[T]:
    allow_empty = allow_empty if default is None else True
    def verify(text: str) -> bool:
        if len(text) < 1 and allow_empty:
            return True
        try:
            convert(text)
            return True
        except (ValueError, ArgumentTypeError):
            return False
    result = select_str(verify, default="" if allow_empty else None, silent=silent)
    return convert(result) if len(result) > 0 or not allow_empty else default





import logging
from typing import Any, Callable, Dict, Generic, List, Literal, Optional, TextIO, Type, TypeVar, Sequence, Union, cast, overload

from . import config
from ..fmt import Color, TextConfig, TextFormat, Text


class ColorFormatter(logging.Formatter):
    """A class for formatting colored logs."""

    def format(self, record: logging.LogRecord) -> str:
        ln = record.levelname.upper()
        text_config = getattr(config, f"TEXT_{ln}_CONFIG", None)
        if text_config is not None:
            record.msg = Text(record.msg, config=text_config).parse()
        return super().format(record)




from enum import Enum


class TextFormat(str,Enum):
    NONE               = 0
    BOLD               = 1
    DARK               = 2
    ITALIC             = 3
    UNDERLINE          = 4
    BLINK              = 5
    INVERSE_BACKGROUND = 7
    INVISIBLE          = 8
    CROSSED            = 9


class Color(int, Enum):
    NONE            = -1
    DARK_GRAY       = 30
    DARK_RED        = 31
    GREEN           = 32
    ORANGE          = 33
    BLUE            = 34
    PURPLE          = 35
    DARK_TURQUOISE  = 36
    LIGHT_GRAY      = 37
    GRAY            = 90
    RED             = 91
    LIME            = 92
    YELLOW          = 93
    LIGHT_BLUE      = 94
    PINK            = 95
    TURQUOISE       = 96
    WHITE           = 97





import os
import sys

from pathlib import Path

from typing import Dict, Callable, Generic, List, Optional, Tuple, Type, TypeVar, cast

import magic

from argparse import ArgumentTypeError

from . import Converter, Validator

def get_mimetype_from_file(file: Path) -> str:
    assert file.is_file()
    return  magic.Magic(mime=True).from_file(file.resolve())


def path_is_readable(path: Path) -> bool:
    return os.access(path, mode=os.R_OK)

def path_is_writable(path: Path) -> bool:  # TODO: handle existing directories
    if path.is_dir():
        raise NotImplementedError
    pre_exist = path.exists()
    if not pre_exist and not path.parent.exists():
        return path_is_writable(path.parent)
    writable = True
    try:
        with path.open("ab") as fobj:
            fobj.flush()
    except IOError:
        writable = False
    if not pre_exist:
        path.unlink()
    return writable

class PathValidator(Validator[Path]):

    def __init__(
            self,
            exists: Optional[bool] = None,
            is_absolute: Optional[bool] = None,
            is_block_device: Optional[bool] = None,
            is_char_device: Optional[bool] = None,
            is_dir: Optional[bool] = None,
            is_fifo: Optional[bool] = None,
            is_file: Optional[bool] = None,
            is_mount: Optional[bool] = None,
            is_relative_to: Optional[Path] = None,
            is_not_relative_to: Optional[Path] = None,
            is_reserved: Optional[bool] = None,
            is_socket: Optional[bool] = None,
            is_symlink: Optional[bool] = None,
            match: Optional[str] = None,
            no_match: Optional[str] = None,
            is_readable: Optional[bool] = None,
            is_writable: Optional[bool] = None,
            mimetype: Optional[str] = None,
            mimetypes: Optional[List[str]] = None) -> None:
        self.exists = exists
        self.is_absolute = is_absolute
        self.is_block_device = is_block_device
        self.is_char_device = is_char_device
        self.is_dir = is_dir
        self.is_fifo = is_fifo
        self.is_file = is_file
        self.is_mount = is_mount
        self.is_relative_to = is_relative_to
        self.is_not_relative_to = is_not_relative_to
        self.is_reserved = is_reserved
        self.is_socket = is_socket
        self.is_symlink = is_symlink
        self.match = match
        self.no_match = no_match
        self.is_readable = is_readable
        self.is_writable = is_writable
        self.mimetypes = mimetypes or []
        if mimetype is not None:
            self.mimetypes.append(mimetype)
        if len(self.mimetypes) > 0:
            assert self.is_file is not False
            self.is_file = True

    def validate(self, value: Path) -> None:
        validators: Dict[Callable[[], bool], Tuple[Optional[bool], str, str]] = {
                value.exists: (self.exists, f"{value} does not exist", f"{value} exists"),
                value.is_absolute: (self.is_absolute, f"{value} is not an absolute path", f"{value} is an absolute path"),
                value.is_block_device: (self.is_block_device, f"{value} is not a block device", f"{value} is a block device"),
                value.is_char_device: (self.is_char_device, f"{value} is not a char device", f"{value} is a char device"),
                value.is_dir: (self.is_dir, f"{value} is not a directory", f"{value} is a directory"),
                value.is_fifo: (self.is_fifo, f"{value} is not a fifo stream", f"{value} is a fifo stream"),
                value.is_file: (self.is_file, f"{value} is not a file", f"{value} is a file"),
                value.is_mount: (self.is_mount, f"{value} is not a mount", f"{value} is a mount"),
                value.is_reserved: (self.is_reserved, f"{value} is not reserved", f"{value} is reserved"),
                value.is_socket: (self.is_socket, f"{value} is not a socket", f"{value} is a socket"),
                value.is_symlink: (self.is_symlink, f"{value} is not a symlink", f"{value} is a symlink"),
                lambda: path_is_readable(value): (self.is_readable, f"{value} is not readable", f"{value} is readable"),
                lambda: path_is_writable(value): (self.is_writable, f"{value} is not writable", f"{value} is writable"),
        }
        for validator, tvalue in validators.items():
            bvalue, vfalse, vtrue = tvalue
            if bvalue is not None:
                if validator() != bvalue:
                    raise ArgumentTypeError(vfalse if bvalue else vtrue)
        if self.is_relative_to is not None:
            if not value.is_relative_to(self.is_relative_to):
                raise ArgumentTypeError(f"{value} is not relative to {self.is_relative_to}")
        if self.is_not_relative_to is not None:
            if value.is_relative_to(self.is_not_relative_to):
                raise ArgumentTypeError(f"{value} is relative to {self.is_not_relative_to}")
        if self.match is not None:
            if not value.match(self.match):
                raise ArgumentTypeError(f"{value} does not match {repr(self.match)}")
        if self.no_match is not None:
            if value.match(self.no_match):
                raise ArgumentTypeError(f"{value} matches {repr(self.no_match)}")
        if len(self.mimetypes) > 0:
            mt = get_mimetype_from_file(value)
            if mt not in self.mimetypes:
                mtl = self.mimetypes.pop()
                if len(self.mimetypes) > 1:
                    mtl = ", ".join(self.mimetypes) + " or " + mtl
                raise ArgumentTypeError(f"{value} is of type {mt}, must be {mtl}")
        return None

class ExtendedPathConverter(Converter[str, Path]):

    def convert(self, value: str) -> Path:
        if value.startswith("~"):
            value = str(Path.home()) + value[1:]
        return Path(value)




import select
import sys
import termios
import tty

from typing import Callable, Optional

from ...fmt import Text, TextConfig
from .. import config

def set_cursor_format(config: Optional[TextConfig] = None) -> None:
    print(Text(" \b", config=config), end="", flush=True)


# read a single character on a unix system
# https://exceptionshub.com/python-read-a-single-character-from-the-user.html
def read_char(validate: Callable[[str], bool] = lambda c: True, silent: bool = False) -> str:
    fd = sys.stdin.fileno()

    old_settings = termios.tcgetattr(fd)
    char: Optional[str] = None
    try:
        tty.setraw(fd)
        while char is None or not validate(char):
            if char is not None:
                set_cursor_format(config.CURSOR_ERROR_FORMAT)
                select.select([sys.stdin], [], [], config.CURSOR_ERROR_TIME)[0]
            set_cursor_format(config.CURSOR_DEFAULT_FORMAT)
            char = sys.stdin.read(1)
            set_cursor_format()
            if ord(char) == 0x3:
                raise KeyboardInterrupt
        if not silent:
            if ord(char) == 0x7f:  # backspace
                print("\b \b", end="", flush=True)
            elif char in ["\r", "\n"]:
                print("\r\n", end="", flush=True)
            else:
                print(char, end="", flush=True)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

    return char


def select_str(verify: Callable[[str], bool], auto_accept: bool = False, silent: bool = False) -> str:
    out = ""
    while True:
        c = read_char(validate=lambda char: (verify(out) and (ord(char) != 0x4 or out=="")) if ord(char) in [0x4, ord("\n"), ord("\r")] else (ord(char) != 0x7f or len(out) > 0), silent=silent)
        ci = ord(c)
        if ci == 0x7f:
            out = out[:-1]
        elif ci in [0x4, 0xd]:  # EOF, Enter
            return out
        else:
            out += c
        if auto_accept and verify(out):
            return out


def select_char(chars: str, ignore_case: bool = True, allow_empty: bool = True, silent: bool = False) -> Optional[str]:
    if ignore_case:
        chars = chars.lower()
    def verify_char(char: str) -> bool:
        test_char = char.lower() if ignore_case else char
        if test_char in chars:
            return True
        if allow_empty and ord(test_char) in [0x4, ord('\r'), ord('\n')]:  # EOF, new line
            return True
        return False
    c = read_char(verify_char, silent=silent)
    if not silent:
        print("\r\n", end="", flush=True)
    return c




from typing import Callable, Dict, Iterable, List, Literal, Optional, Sequence, Set, TypeVar, Union, cast, overload

from ..fmt import Text
from . import inpt, out

T = TypeVar("T")

@overload
def get_bool(text: Union[str, Text], default: bool, inline: Optional[bool] = None) -> bool:
    ...

@overload
def get_bool(text: Union[str, Text], default: Optional[bool] = None, inline: Optional[bool] = None) -> Optional[bool]:
    ...

def get_bool(text: Union[str, Text], default: Optional[bool] = None, inline: Optional[bool] = None) -> Optional[bool]:
    info = "y/n"
    if default is not None:
        info = "Y/n" if default else "y/N"
    out.print_question(Text(text, " [", info, "]"), inline=inline)
    return inpt.select_bool(default)


@overload
def get_str(text: Union[str, Text], verify: Callable[[str], bool], default: str, silent: bool = False, inline: Optional[bool] = None) -> str:
    ...

@overload
def get_str(text: Union[str, Text], verify: Callable[[str], bool] = lambda text: True, default: Optional[str] = None, silent: bool = False, inline: Optional[bool] = None) -> Optional[str]:
    ...

def get_str(text: Union[str, Text], verify: Callable[[str], bool] = lambda text: True, default: Optional[str] = None, silent: bool = False, inline: Optional[bool] = None) -> Optional[str]:
    def_text: Union[str, Text] = ""
    if default is not None:
        def_text = Text(" [", default, "]")
    out.print_question(Text(text, def_text), inline=inline)
    return inpt.select_str(verify, default, silent)


@overload
def get(text: Union[str, Text], convert: Callable[[str], T], default: T, default_str: Optional[Union[str, Text]] = None, allow_empty: bool = False, silent: bool = False, inline: Optional[bool] = None) -> T:
    ...

@overload
def get(text: Union[str, Text], convert: Callable[[str], T], default: Optional[T] = None, default_str: Optional[Union[str, Text]] = None, allow_empty: Literal[False] = False, silent: bool = False, inline: Optional[bool] = None) -> T:
    ...

@overload
def get(text: Union[str, Text], convert: Callable[[str], T], default: Optional[T] = None, default_str: Optional[Union[str, Text]] = None, allow_empty: bool = False, silent: bool = False, inline: Optional[bool] = None) -> Optional[T]:
    ...

def get(text: Union[str, Text], convert: Callable[[str], T], default: Optional[T] = None, default_str: Optional[Union[str, Text]] = None, allow_empty: bool = False, silent: bool = False, inline: Optional[bool] = None) -> Optional[T]:
    def_text: Union[str, Text] = ""
    if default is not None:
        def_text = Text(" [", default_str or str(default), "]")
    out.print_question(Text(text, def_text), inline=inline)
    return inpt.select(convert, default, allow_empty, silent)


import math, re, sys


@overload
def ask_nlist(
    text: str,
    seq_items: Iterable[T],
    allow_range: Literal[False] = False,
    topiced_items: Optional[Union[Dict[str, List[T]], Dict[Optional[str], List[T]], Dict[str, Sequence[T]], Dict[Optional[str], Sequence[T]]]] = None,
    custom_keys: Literal[None] = None,
    custom_all_key: Literal[None] = None,
    default: Optional[T] = None,
    default_index: Optional[int] = None,
    display: Callable[[T], str] = str,
    skip_if_one: bool = False,
) -> T:
    ...


@overload
def ask_nlist(
    text: str,
    seq_items: Iterable[T],
    allow_range: Literal[True],
    topiced_items: Optional[Union[Dict[str, List[T]], Dict[Optional[str], List[T]], Dict[str, Sequence[T]], Dict[Optional[str], Sequence[T]]]] = None,
    custom_keys: Literal[None] = None,
    custom_all_key: Optional[str] = None,
    default: Optional[T] = None,
    default_index: Optional[int] = None,
    display: Callable[[T], str] = str,
    skip_if_one: bool = False,
) -> List[T]:
    ...


@overload
def ask_nlist(
    text: str,
    seq_items: Iterable[T],
    allow_range: Literal[False] = False,
    topiced_items: Optional[Union[Dict[str, List[T]], Dict[Optional[str], List[T]], Dict[str, Sequence[T]], Dict[Optional[str], Sequence[T]]]] = None,
    custom_keys: Optional[Dict[str, str]] = None,
    custom_all_key: Literal[None] = None,
    default: Optional[Union[T, str]] = None,
    default_index: Optional[Union[int, str]] = None,
    display: Callable[[T], str] = str,
    skip_if_one: bool = False,
) -> Union[str, T]:
    ...


@overload
def ask_nlist(
    text: str,
    seq_items: Iterable[T],
    allow_range: Literal[True],
    topiced_items: Optional[Union[Dict[str, List[T]], Dict[Optional[str], List[T]], Dict[str, Sequence[T]], Dict[Optional[str], Sequence[T]]]] = None,
    custom_keys: Optional[Dict[str, str]] = None,
    custom_all_key: Optional[str] = None,
    default: Optional[Union[T, str]] = None,
    default_index: Optional[Union[int, str]] = None,
    display: Callable[[T], str] = str,
    skip_if_one: bool = False,
) -> Union[str, List[T]]:
    ...


@overload
def ask_nlist(
    text: str,
    seq_items: Iterable[T],
    allow_range: bool = False,
    topiced_items: Optional[Union[Dict[str, List[T]], Dict[Optional[str], List[T]], Dict[str, Sequence[T]], Dict[Optional[str], Sequence[T]]]] = None,
    custom_keys: Optional[Dict[str, str]] = None,
    custom_all_key: Optional[str] = None,
    default: Optional[Union[T, str]] = None,
    default_index: Optional[Union[int, str]] = None,
    display: Callable[[T], str] = str,
    skip_if_one: bool = False,
) -> Union[str, T, List[T]]:
    ...


# ask using a list of valid items
# if the index of the selected item is required, enumerate(items) can be passed as arguments
def ask_nlist(
    text: str,
    seq_items: Iterable[T],
    allow_range: bool = False,
    topiced_items: Optional[Union[Dict[str, List[T]], Dict[Optional[str], List[T]], Dict[str, Sequence[T]], Dict[Optional[str], Sequence[T]]]] = None,
    custom_keys: Optional[Dict[str, str]] = None,
    custom_all_key: Optional[str] = None,
    default: Optional[Union[T, str]] = None,
    default_index: Optional[Union[int, str]] = None,
    display: Callable[[T], str] = str,
    skip_if_one: bool = False,
) -> Union[str, T, List[T]]:
    if default is not None and default_index is not None:
        assert False

    if topiced_items is None:
        topiced_items = cast(Union[Dict[str, List[T]], Dict[Optional[str], List[T]]], {})
    items = list(seq_items)
    if None in topiced_items:  # topiced_items key has Optional[str] type
        items += list(topiced_items[None])  # type: ignore[index]  # mypy cannot recognize None as keytype
    topiced_items = {topic: entries for topic, entries in topiced_items.items() if topic is not None}
    if custom_keys:
        for key in custom_keys:
            if re.match(r"^(\d+|\*)$", key):
                if key == "*":
                    if custom_all_key is not None:
                        raise Exception("custom key must not be '*'")
                else:
                    raise Exception("custom key must not be a number")
    else:
        custom_keys = {}

    if custom_all_key is not None:
        custom_keys = {**custom_keys, "*": custom_all_key}

    # sort dicts
    sorted_topiced_items = {topic: topiced_items[topic] for topic in sorted(topiced_items)}
    sorted_custom_keys = {key: custom_keys[key] for key in sorted(custom_keys)}

    result_list: List[T] = []
    for topic_items in sorted_topiced_items.values():
        result_list.extend(topic_items)
    result_list.extend(items)
    entry_count = len(result_list)

    default_str = None
    default_custom_key = None
    if default is not None:
        if default in result_list:  # implies 'default' is of type T
            # mypy cannot recognise implied type
            default_str = str(result_list.index(default))  # type: ignore[arg-type]
            default_custom_key = False
        elif isinstance(default, str) and default in custom_keys.values():
            default_str = {v: k for k, v in custom_keys.items()}[default]
            default_custom_key = True
        else:
            assert False
    elif default_index is not None:
        if isinstance(default_index, int):
            if 1 <= default_index <= entry_count:
                default_str = str(default_index)
                default_custom_key = False
                default = result_list[default_index-1]
            else:
                assert False
        elif default_index in custom_keys:
            default_str = default_index
            default_custom_key = True
            default = custom_keys[default_index]
        else:
            assert False

    if entry_count == 1 and skip_if_one:
        if allow_range:
            return [result_list[0]]
        return result_list[0]

    max_key_length = max(0, int(math.log10(max(entry_count, 1))+1), *[len(key) for key in custom_keys.keys()])

    fmt_str = f" [%{max_key_length}s]: %s"

    def print_choices() -> None:
        print(f">>> {text}", flush=True)
        item_id = 1
        for header, entry_items in sorted_topiced_items.items():
            print(header)
            for item in entry_items:
                print(fmt_str % (item_id, display(item)))
                item_id += 1
            print()
        for item in items:
            print(fmt_str % (item_id, display(item)))
            item_id += 1
        for key, value in sorted_custom_keys.items():
            print(fmt_str % (key, value))

    print_choices()
    from .inpt.selector import select_str
    def vrfy(choice_str: str) -> bool:
        if default is not None and choice_str == "":
            return True

        if custom_all_key is not None and choice_str == "*":
            return True
        if choice_str in sorted_custom_keys:
            return True
        if allow_range:
            choices: Set[int] = set()
            invalid_choices = []

            for range_str in map(str.strip, choice_str.split(",")):
                range_match = re.match(r"^(\d+|(\d+)-(\d+))$", range_str)
                if range_match:  # range / number
                    if range_match.group(2) and range_match.group(3):  # range
                        start = int(range_match.group(2))
                        end = int(range_match.group(3))
                        if 1 <= start <= end <= entry_count:
                            choices.update(range(start, end+1))
                            continue
                    elif 1 <= int(range_match.group(1)) <= entry_count:
                        choices.add(int(range_match.group(1)))
                        continue
                invalid_choices.append(range_str)

            return not invalid_choices

        return choice_str.isdigit() and (int(choice_str) > 0) and (int(choice_str) <= len(result_list))

    print_choices()
    print("> ", end="", flush=True)
    choice_str = select_str(verify=vrfy)

    if default is not None and choice_str == "":
        if allow_range and default_custom_key is False:
            return [default]  # type: ignore[list-item]  # default_custom_key enforces type T
        return default

    if custom_all_key is not None and choice_str == "*":
        return result_list
    if choice_str in custom_keys:
        return choice_str
    if allow_range:
        choices: Set[int] = set()
        invalid_choices = []

        for range_str in map(str.strip, choice_str.split(",")):
            range_match = re.match(r"^(\d+|(\d+)-(\d+))$", range_str)
            if range_match:  # range / number
                if range_match.group(2) and range_match.group(3):  # range
                    start = int(range_match.group(2))
                    end = int(range_match.group(3))
                    if 1 <= start <= end <= entry_count:
                        choices.update(range(start, end+1))
                        continue
                elif 1 <= int(range_match.group(1)) <= entry_count:
                    choices.add(int(range_match.group(1)))
                    continue
            invalid_choices.append(range_str)

        if invalid_choices:
            pass # impossible
        return [result_list[choice-1] for choice in choices]

    if not choice_str.isdigit() or (int(choice_str) < 1) or (int(choice_str) > len(result_list)):
        #print_error("Invalid value \"%s\". Please enter again" % choice_str)
        pass  # impossible

    return result_list[int(choice_str)-1]




import sys

from typing import Callable, Generic, Optional, Type, TypeVar, cast
if sys.version_info >= (3, 8):
    from typing import get_args

from argparse import ArgumentTypeError


N = TypeVar("N", int, float)


class Range(Generic[N]):

    @property
    def type(self) -> Type[N]:
        if self._type is not None:
            return self._type
        t = getattr(self, "__orig_class__", None)
        assert t is not None, f"{self.__class__.__name__} must overwrite property type with correct type"
        if sys.version_info >= (3, 8):
            return cast(Type[N], get_args(t)[0])
        else:
            return t.__args__[0]  # type: ignore[no-any-return]

    def __init__(
            self,
            minv: Optional[N] = None,
            maxv: Optional[N] = None,
            min_eq: bool = True,
            max_eq: bool = True,
            typ: Optional[Type[N]] = None) -> None:
        self.min: Optional[N] = minv
        self.max: Optional[N] = maxv
        self.min_eq = min_eq
        self.max_eq = max_eq
        self._type: Optional[Type[N]] = typ

    def __call__(self, value: str) -> N:
        try:
            parg = self.type(value)
        except ValueError as e:
            raise ArgumentTypeError("value must be of type %s" % self.type.__name__) from e
        bmin = self.min is None or self.min < parg or (self.min_eq and parg == self.min)
        bmax = self.max is None or self.max > parg or (self.max_eq and parg == self.max)
        if bmax and bmin:
            return parg
        if self.min is not None:
            if self.max is not None:
                raise ArgumentTypeError("value must be in range %s to %s" % (self.min, self.max))
            if self.min == 0:
                raise ArgumentTypeError("value must be positive")
            if self.min_eq:
                raise ArgumentTypeError("value must be at least %s" % self.min)
            raise ArgumentTypeError("value must be greater than %s" % self.min)
        if self.max_eq:
            raise ArgumentTypeError("value must be at most %s" % self.max)
        if self.max == 0:
            raise ArgumentTypeError("value must be negative")
        raise ArgumentTypeError("value must be less than %s" % self.max)


class RangeInt(Range[int]):

    @property
    def type(self) -> Type[int]:
        return int


class RangeFloat(Range[float]):

    @property
    def type(self) -> Type[float]:
        return float




import sys
from enum import Enum
from typing import Generic, Optional, Type, TypeVar, cast
if sys.version_info >= (3, 8):
    from typing import get_args

from argparse import ArgumentTypeError


E = TypeVar("E", bound=Enum)


class EnumSelector(Generic[E]):

    @property
    def type(self) -> Type[E]:
        if self._type is not None:
            return self._type
        t = getattr(self, "__orig_class__", None)
        assert t is not None, f"{self.__class__.__name__} must overwrite property type with correct type"
        if sys.version_info >= (3, 8):
            return cast(Type[E], get_args(t)[0])
        else:
            return t.__args__[0]

    def __init__(self, typ: Optional[Type[E]] = None) -> None:
        self._type: Optional[Type[E]] = typ

    def __call__(self, value: str) -> E:
        args = {str(entry): entry for entry in self.type}
        if value in args:
            return args[value]
        raise ArgumentTypeError(f"invalid {self.type.__name__} value: '{value}'")




import inspect
import sys

from abc import ABC, ABCMeta, abstractmethod

from typing import Any, Callable, Generic, List, Optional, Type, TypeVar, Union

from argparse import ArgumentTypeError

T = TypeVar("T")
I = TypeVar("I")
O = TypeVar("O")
O1 = TypeVar("O1")
O2 = TypeVar("O2")


class Converter(ABC, Generic[I, O]):

    @abstractmethod
    def convert(self, value: I) -> O:
        raise NotImplementedError

    def __call__(self, value: I) -> O:
        return self.convert(value)

    def is_valid(self, value: I) -> bool:
        try:
            self.convert(value)
        except (ValueError, ArgumentTypeError):
            return False
        return True

    def to_validator(self) -> "ConverterValidator[I]":
        return ConverterValidator(self)

    def chain(self, other: "Union[Converter[O, T], Callable[[O], T]]") -> "ConverterChain[I, T]":
        if not isinstance(other, Converter):
            other = FunctionConverter[O, T](other)
        return ConverterChain(self, other)

    def rchain(self, other: "Union[Converter[T, I], Callable[[T], I]]") -> "ConverterChain[T, O]":
        if not isinstance(other, Converter):
            other = FunctionConverter(other)
        return other.chain(self)

    def branch(self, other: "Union[Converter[I, T], Callable[[I], T]]") -> "ConverterBranch[I, O, T]":
        if not isinstance(other, Converter):
            other = FunctionConverter(other)
        return ConverterBranch(self, other)

    def rbranch(self, other: "Union[Converter[I, T], Callable[[I], T]]") -> "ConverterBranch[I, T, O]":
        if not isinstance(other, Converter):
            other = FunctionConverter(other)
        return other.branch(self)

    def __truediv__(self, other: "Union[Converter[O, T], Callable[[O], T], Any]") -> "ConverterChain[I, T]":
        if not callable(other):
            return NotImplemented
        return self.chain(other)

    def __rtruediv__(self, other: "Union[Callable[[T], I], Any]") -> "ConverterChain[T, O]":
        if not callable(other):
            return NotImplemented
        return self.rchain(other)

    def __rshift__(self, other: "Union[Converter[O, T], Callable[[O], T], Any]") -> "ConverterChain[I, T]":
        return self / other

    def __rrshift__(self, other: "Union[Callable[[T], I], Any]") -> "ConverterChain[T, O]":
        return other / self

    def __mul__(self, other: "Union[Converter[O, T], Callable[[O], T], Any]") -> "ConverterChain[I, T]":
        return self / other

    def __rmul__(self, other: "Union[Callable[[T], I], Any]") -> "ConverterChain[T, O]":
        return other / self

    def __add__(self, other: "Union[Converter[I, T], Callable[[I], T], Any]") -> "ConverterBranch[I, O, T]":
        if not callable(other):
            return NotImplemented
        return self.branch(other)

    def __radd__(self, other: "Union[Converter[I, T], Callable[[I], T], Any]") -> "ConverterBranch[I, T, O]":
        if not callable(other):
            return NotImplemented
        return self.rbranch(other)

    def __invert__(self) -> "ConverterValidator[I]":
        return self.to_validator()

class Validator(Converter[T, T]):

    def __init__(self, error_msg: Optional[str] = None) -> None:
        super().__init__()
        self.error_msg = error_msg  # TODO: check that error_msg is valid format str

    def convert(self, value: T) -> T:
        if self.validate(value) is False:
            msg = "Validation of %s failed"
            if self.error_msg is not None:
                try:
                    msg = self.error_msg % value
                except ValueError:
                    msg = self.error_msg
            raise ArgumentTypeError(msg % value)
        return value

    @abstractmethod
    def validate(self, value: T) -> Optional[bool]:
        raise NotImplementedError

class FunctionConverter(Converter[I, O]):

    def __init__(self, convert: Callable[[I], O]) -> None:
        super().__init__()
        self._converter = convert

    def convert(self, value: I) -> O:
        return self._converter(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self._converter)})"

class FunctionValidator(Validator[T]):

    def __init__(self, validate: Callable[[T], Optional[bool]], error_msg: Optional[str] = None) -> None:
        super().__init__(error_msg)
        self._validator = validate

    def validate(self, value: T) -> Optional[bool]:
        return self._validator(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self._validator)})"

class ConverterValidator(Validator[T]):

    def __init__(self, converter: Converter[T, O]) -> None:
        super().__init__()
        self._converter = converter

    def validate(self, value: T) -> None:
        self._converter(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self._converter)})"

class ConverterChain(FunctionConverter[I, O]):

    def __init__(self, first_converter: Converter[I, T], second_converter: Converter[T, O]) -> None:
        super().__init__(lambda value: second_converter.convert(first_converter.convert(value)))
        self.first_converter = first_converter
        self.second_converter = second_converter

    def __repr__(self) -> str:
        converter = []
        stack = [self.second_converter, self.first_converter]
        while len(stack) > 0:
            c = stack.pop()
            if isinstance(c, ConverterChain):
                stack.append(c.second_converter)
                stack.append(c.first_converter)
            else:
                converter.append(c)
        return f"<ConverterChain: {', '.join(repr(c) for c in converter)}>"

class ConvertError(ArgumentTypeError):
    
    def __init__(self, msg: str, *errors: Union[ValueError, ArgumentTypeError]) -> None:
        super().__init__(msg)
        self.errors = errors

class ConverterBranch(Converter[I, Union[O1, O2]]):

    def __init__(self, first_converter: Converter[I, O1], second_converter: Converter[I, O2], error_msg: str = "could not convert value '%s'") -> None:
        self.error_msg = error_msg
        self.first_converter = first_converter
        self.second_converter = second_converter

    def convert(self, value: I) -> Union[O1, O2]:
        err_list: List[Union[ValueError, ArgumentTypeError]] = []
        try:
            return self.first_converter(value)
        except ConvertError as cerr:
            err_list.extend(cerr.errors)
        except (ValueError, ArgumentTypeError) as err:
            err_list.append(err)
        try:
            return self.second_converter(value)
        except ConvertError as cerr:
            err_list.extend(cerr.errors)
        except (ValueError, ArgumentTypeError) as err:
            err_list.append(err)
        try:
            msg = self.error_msg % value
        except TypeError:
            msg = self.error_msg
        raise ConvertError(msg, *err_list)


    def __repr__(self) -> str:
        converter = []
        stack = [self.second_converter, self.first_converter]
        while len(stack) > 0:
            c = stack.pop()
            if isinstance(c, ConverterChain):
                stack.append(c.second_converter)
                stack.append(c.first_converter)
            else:
                converter.append(c)
        return f"<ConverterBranch: {', '.join(repr(c) for c in converter)}>"



from ..fmt import Color, TextConfig

TEXT_DEBUG_CONFIG = TextConfig(Color.GRAY)
TEXT_INFO_CONFIG = TextConfig(Color.WHITE)
TEXT_WARNING_CONFIG = TextConfig(Color.YELLOW)
TEXT_ERROR_CONFIG = TextConfig(Color.DARK_RED)
TEXT_CRITICAL_CONFIG = TextConfig(Color.DARK_RED)

EXCEPTION_INFORMATION = {
    NotImplementedError: "this feature is not implemented",
}



from ..fmt import Color, TextConfig, TextFormat

QUESTION_FORMAT = TextConfig(color=Color.PURPLE)

CURSOR_DEFAULT_FORMAT = TextConfig(color=Color.NONE, color_bg=Color.NONE, attrs=[TextFormat.DARK])
CURSOR_ERROR_FORMAT   = TextConfig(color=Color.RED, color_bg=Color.NONE, attrs=[TextFormat.DARK])

CURSOR_ERROR_TIME = 0.6

AUTO_INLINE_MAX_CHARACTER_COUNT = 30

QUESTION_PREFIX = ">>> "
QUESTION_INLINE_SUFFIX = ": "



from typing import Any, List, Optional, Union

from . import Color, TextFormat

class TextConfig:

    def __init__(self, color: Optional[Color] = None, color_bg: Optional[Color] = None, attrs: Optional[List[TextFormat]] = None) -> None:
        self.color = color
        self.color_bg = color_bg
        self.attrs = attrs or []

    def formatter(self) -> str:
        fmtr: List[str] = []
        if self.attrs:
            fmtr.extend(str(attr.value) for attr in self.attrs)
        if self.color is not None and self.color != Color.NONE:
            fmtr.append("%d" % self.color)
        if self.color_bg is not None and self.color_bg != Color.NONE:
            fmtr.append("%d" % (self.color_bg + 10))
        fmt = f"\033[{';'.join(fmtr)}m"
        return "\033[m" + fmt

    def append(self, other: "TextConfig") -> "TextConfig":
        new_attrs = self.attrs + other.attrs
        color = other.color if other.color is not None else self.color
        color_bg = other.color_bg if other.color_bg is not None else self.color_bg
        return TextConfig(color, color_bg, new_attrs)

    def is_empty(self) -> bool:
        return not self.color and not self.color_bg and not self.attrs

class Text:

    def __init__(self, *snippets: Union["Text", str], color: Optional[Color] = None, color_bg: Optional[Color] = None, attrs: Optional[List[TextFormat]] = None, config: Optional[TextConfig] = None) -> None:
        self.snippets = snippets
        self.config = TextConfig(color, color_bg, attrs)
        if config:
            assert self.config.is_empty()
            self.config = config

    def parse(self, *args: Any, **kwargs: Any) -> str:
        return self._parse(TextConfig(), *args, **kwargs)

    @property
    def unformatted(self) -> str:
        return "".join(snip.unformatted if isinstance(snip, Text) else str(snip) for snip in self.snippets)

    def _parse(self, config: TextConfig, *args: Any, **kwargs: Any) -> str:
        full_config = config.append(self.config)
        fmt = full_config.formatter()
        output: str = fmt
        for snip in self.snippets:
            if isinstance(snip, Text):
                output += snip._parse(full_config, *args, **kwargs)
                output += fmt
            else:
                output += str(snip)
        return output + config.formatter()

    def __str__(self) -> str:
        return self.parse()




from .validator import Converter, Validator, FunctionConverter, FunctionValidator
from .number import Range, RangeFloat, RangeInt
from .path import ExtendedPathConverter, PathValidator
from .selector import EnumSelector
